/*
 * mcp4725.c
 *
 * Simple MCP725 12b I2C DAC Driver
 *
 * Created: 5. 6. 2023 10:51:34
 *  Author: greif
 */ 

#include "mcp4725.h"
#include "i2cmaster.h" // Standard I2C library



#define DAC_REG_WRITE 0x40 // Write mode - DAC register only, EEPROM unchanged
#define DAC_EEPROM_REG_WRITE 0x60 // // Write mode - both DAC register and EEPROM
#define DAC_RESET 0x06 // MCP4725 reset command (general call)



// Perform internal MCP4725 reset - EEPROM data uploaded to the DAC register
void resetDAC(){
	i2c_start_wait( (DAC_ADDR << 1) | I2C_WRITE); // Start I2C communication with MCP4725 (ack polling), write mode
	i2c_write(0x00); // Start address 0x00 (I2C general call)
	i2c_write(DAC_RESET); // Send reset command
	i2c_stop(); // Stop I2C communication
}



// Initialize MCP4725
void DACInit(){
	i2c_init(); // initialize the I2C library
	resetDAC(); // Issue a general call reset after power-on (recommended by manufacturer - in case the automatic power-on reset did not work)
}



/* Write MCP4725 register
*
* Parameters: 
*	const uint16_t value - 12b value representing output voltage
*	const int EEPROM - write mode:
*		if true, value is written to both DAC register and EEPROM - slow
*		if false, value is written only to the DAC register - fast
*/
void writeDAC(const uint16_t value, const int EEPROM){
	i2c_start_wait( (DAC_ADDR << 1) | I2C_WRITE); // Start I2C communication with MCP4725 (ack polling), write mode
	i2c_write(EEPROM ? DAC_EEPROM_REG_WRITE : DAC_REG_WRITE); // Set write address depending on write mode (parameter)
	i2c_write((value & 0xFF0) >> 4); // Write upper data bits (D11, D10, D9, D8, D7, D6, D5, D4)
	i2c_write((value & 0xF) << 4); // Write Lower data bits (D3, D2, D1, D0)
	i2c_stop(); // Stop I2C communication
}



/* Read MCP4725 register
*
* Parameters:
*	uint8_t *buffer - buffer pointer to store read bits
*	const uint8_t length - buffer length
*/
void readDACRegister(uint8_t *buffer, const uint8_t length){
	i2c_start_wait( (DAC_ADDR << 1) | I2C_WRITE); // Start I2C communication with MCP4725 (ack polling), write mode
	i2c_write(0x00); // Set the address to the start (0x00)
	
	i2c_rep_start( (DAC_ADDR << 1) | I2C_READ ); // Start I2C communication with MCP4725 again, this time in read mode
	
	// Read (length-1) bits from the register
	uint8_t counter = 0;
	while (counter < length-1){
		buffer[counter++] = i2c_readAck(); // Read the bit, continue requesting data
	}
	buffer[counter] = i2c_readNak(); // Read the last wanted bit, do not request any other data
	
	i2c_stop(); // Stop I2C communication
}



// Check if writing to the EEPROM has finished
int DACReady(){
	uint8_t buffer[1];
	readDACRegister(buffer, 1); // Read the first byte
	return ((buffer[0] & 0x80) > 0); // Return the status of the first bit
}



// Read the set value form MCP4725 EEPROM
uint16_t readEEPROM(){
	while(!DACReady()); // Wait until writing to the EEPROM is finished
	
	uint8_t buffer[5];
	readDACRegister(buffer, 5); // Read 5 bytes (EEPROM data in the 4-5th byte)
	
	// Convert read bytes to the actual set 12b value
	uint16_t result = buffer[3] & 0x0F;
	result = result << 8;
	result = result + buffer[4];
	
	return result;
}



// Read the set value from MCP4725 DAC register
uint16_t readDACValue(){
	while(!DACReady()); // Wait until writing to the EEPROM is finished
	
	uint8_t buffer[3];
	readDACRegister(buffer, 3); // Read 3 bytes (DAC data in the 2-3th byte)
	
	// Convert read bytes to the actual set 12b value
	uint16_t result = buffer[1];
	result = result << 4;
	result = result + (buffer[2] >> 4);
	
	return result;
}



/* Set MCP4725 value
* Wrapper of void writeDAC(const uint16_t value, const int EEPROM)
* Added input range check
*
* Parameters:
*	const uint16_t value - 12b value representing output voltage
*	const int persistent - write mode:
*		if true, value is written to both DAC register and EEPROM - slow, value stays unchanged after reset
*		if false, value is written only to the DAC register - fast, value changes after reset - to the EEPROM value
*
* Return:
*	1 if value out of range
*	0 if write successful
*/
int setDACValue(uint16_t value, const int persistent){
	// Check the input range
	if(value < DAC_MIN || value > DAC_MAX){
		return 1;
	}
	
	while(!DACReady()); // Wait until writing to the EEPROM is finished
	writeDAC(value, persistent); // Depending on the write mode, write the value to MCP4725
	
	return 0;
}